package com.jclark.xsl;

import java.util.Enumeration;
import java.util.Vector;

public class PatternParser {
  private PatternParser() { }

  public static
  Enumeration createMatchers(Node node, String pattern) throws XSLException {
    Vector v = new Vector();
    PatternTokenizer tokens = new PatternTokenizer(node, pattern);
    do {
      v.addElement(parseMatcher(tokens, null));
    } while ("|".equals(tokens.getCurrent()));
    if (tokens.getCurrent() != null)
      throw new XSLException("unexpected token \"" + tokens.getCurrent() + "\"",
			     node);
    return v.elements();
  }

  public static
  Matchable createMatchable(Node node, String pattern) throws XSLException {
    Matchable m = null;
    PatternTokenizer tokens = new PatternTokenizer(node, pattern);
    do {
      if (m == null)
	m = parseMatcher(tokens, null);
      else
	m = new OrPattern(parseMatcher(tokens, null), m);
    } while ("|".equals(tokens.getCurrent()));
    if (tokens.getCurrent() != null)
      throw new XSLException("unexpected token \"" + tokens.getCurrent() + "\"",
			     node);
    return m;
  }

  public static
  int findExpressionEnd(String value, int i) {
    int valueLen = value.length();
    char quote = '\0';
    for (; i < valueLen; i++) {
      char c = value.charAt(i);
      switch (c) {
      case '}':
	if (quote == '\0')
	  return i;
	break;
      case '\"':
      case '\'':
	if (quote == c)
	  quote = '\0';
	else if (quote == '\0')
	  quote = c;
        break;
      }
    }
    return -1;
  }

  public static
  Expression createAttributeValueTemplate(Node node, String value) throws XSLException {
    Expression prev = null;
    StringBuffer buf = new StringBuffer();
    int valueLen = value.length();
    for (int i = 0; i < valueLen; i++) {
      char c = value.charAt(i);
      switch (c) {
      case '{':
	if (i + 1 < valueLen && value.charAt(i + 1) == '{') {
	  i++;
	  buf.append('{');
	}
	else {
	  int n = findExpressionEnd(value, i + 1);
	  if (n < 0)
	    throw new XSLException("missing }", node);
	  Expression expr = createExpression(node, value.substring(i + 1, n));
	  if (buf.length() > 0) {
	    if (prev == null)
	      prev = new ConstantExpression(buf.toString());
	    else
	      prev = new AppendExpression(prev, new ConstantExpression(buf.toString()));
	    buf.setLength(0);
	  }
	  if (prev == null)
	    prev = expr;
	  else
	    prev = new AppendExpression(prev, expr);
	  i = n;
	}
	break;
      case '}':
	buf.append('}');
	if (i + 1 < valueLen && value.charAt(i + 1) == '}')
	  i++;
	break;
      default:
	buf.append(c);
	break;
      }
    }
    if (buf.length() > 0) {
      if (prev == null)
  	return new ConstantExpression(buf.toString());
      else
	return new AppendExpression(prev, new ConstantExpression(buf.toString()));
    }
    if (prev != null)
      return prev;
    return new ConstantExpression("");
  }

  public static
  Expression createExpression(Node node, String pattern) throws XSLException {
    PatternTokenizer tokens = new PatternTokenizer(node, pattern);
    String t = tokens.next();
    if ("constant(".equals(t)) {
      String name = tokens.next();
      if (name == null || !PatternTokenizer.isName(name))
	throw new XSLException("expected name", node);
      if (!")".equals(tokens.next()))
	throw new XSLException("expected \")\"", node);
      if (tokens.next() != null)
	throw new XSLException("junk after constant expression", node);
      return new ConstantValueExpression(tokens.expandElementTypeName(name));
    }
    if ("arg(".equals(t)) {
      String name = tokens.next();
      if (name == null || !PatternTokenizer.isName(name))
	throw new XSLException("expected name", node);
      if (!")".equals(tokens.next()))
	throw new XSLException("expected \")\"", node);
      if (tokens.next() != null)
	throw new XSLException("junk after macro arg expression", node);
      return new MacroArgExpression(tokens.expandAttributeName(name));
    }
    return new QueryExpression(createQuery(node, pattern));
  }

  public static
  Query createQuery(Node node, String pattern) throws XSLException {
    PatternTokenizer tokens = new PatternTokenizer(node, pattern);
    SubQuery query = parseQuery(tokens, true);
    while (tokens.getCurrent() != null) {
      SubQuery alternate = parseQuery(tokens, true);
      SubQuery u = query.createUnion(alternate);
      if (u != null)
	query = u;
      else
	query = new UnionQuery(query, alternate);
    }
    return query;
  }

  private static SubQuery parseQuery(PatternTokenizer tokens, boolean isFirst) throws XSLException {
    tokens.next();
    return parseQueryCurrent(tokens, isFirst);
  }

  private static
  SubQuery parseQueryCurrent(PatternTokenizer tokens, boolean isFirst) throws XSLException {
    String head = tokens.getCurrent();
    if (head == null
	|| (!PatternTokenizer.isName(head)
	    && !head.equals("*")
	    && !head.equals("..")
	    && !head.equals("attribute(")
	    && !(isFirst && (head.equals(".")
			     || head.equals("/")
			     || head.equals("id(")
			     || head.equals("ancestor(")))))
      throw new XSLException("expected element name or wildcard or anchor",
			     tokens.getNode());
    String t;
    SubQuery query;
    if (head.equals("/")) {
      t = tokens.next();
      if (t == null || t.equals("|"))
	return new RootQuery();
      return new ComposeQuery(new RootQuery(), parseQueryCurrent(tokens, false));
    }
    if (head.equals(".")) {
      t = tokens.next();
      query = new SelfQuery();
    }
    else if (head.equals("..")) {
      t = tokens.next();
      query = new ParentQuery();
    }
    else if (head.equals("id(")) {
      t = tokens.next();
      if (!PatternTokenizer.isName(t))
	throw new XSLException("expected id", tokens.getNode());
      query = new IdQuery(t);
      if (!")".equals(tokens.next()))
	throw new XSLException("expected )", tokens.getNode());
      t = tokens.next();
    }
    else if (head.equals("ancestor(")) {
      Matcher matcher = parseMatcher(tokens, null);
      if (!")".equals(tokens.getCurrent()))
	throw new XSLException("expected )", tokens.getNode());
      query = new AncestorQuery(matcher);
      t = tokens.next();
    }
    else if (head.equals("attribute(")) {
      t = tokens.next();
      if (!PatternTokenizer.isName(t))
	throw new XSLException("expected attribute name", tokens.getNode());
      query = new AttributePattern(tokens.expandAttributeName(t), null);
      t = tokens.next();
      if (!")".equals(t))
	throw new XSLException("expected )", tokens.getNode());
      t = tokens.next();
    }
    else {
      Qualifier qualifier = null;
      t = tokens.next();
      if ("[".equals(t)) {
	qualifier = parseQualifiers(tokens, null);
	t = tokens.next();
      }
      if (head.equals("*"))
	query = new WildcardPattern(qualifier);
      else
	query = new ElementPattern(tokens.expandElementTypeName(head), qualifier);
    }
    if (t == null)
      return query;
    if (t.equals("/"))
      return new ComposeQuery(query, parseQuery(tokens, false));
    if (t.equals("//")) {
      Matcher pattern = parseMatcher(tokens, null);
      if (pattern instanceof AttributePattern)
	return new ComposeQuery(query,
				new AttributeDescendantsQuery((AttributePattern)pattern));
      else
	return new ComposeQuery(query,
				new DescendantsQuery(pattern));
    }
    if (t.equals("|"))
      return query;
    throw new XSLException("expected / or // or |", tokens.getNode());
  }


  private static
  Matcher parseMatcher(PatternTokenizer tokens, Qualifier qualifier) throws XSLException {
    tokens.next();
    return parseMatcherCurrent(tokens, qualifier);
  }

  private static
  Matcher parseMatcherCurrent(PatternTokenizer tokens, Qualifier qualifier) throws XSLException {
    String head = tokens.getCurrent();
    if (head == null
	|| (!PatternTokenizer.isName(head)
	    && !head.equals("*")
	    && !head.equals("attribute(")
	    && !(head.equals("/") && qualifier == null)))
      throw new XSLException("expected element name or wildcard", tokens.getNode());
    String t = tokens.next();
    if (head.equals("/")) {
      if (t == null || t.equals(")") || t.equals("|"))
	return new RootMatcher();
      return parseMatcherCurrent(tokens,
				 new ParentQualifier(new RootMatcher(), null));
    }
    Matcher pattern;
    if (head.equals("attribute(")) {
      if (!PatternTokenizer.isName(t))
	throw new XSLException("expected attribute name", tokens.getNode());
      pattern = new AttributePattern(tokens.expandAttributeName(t), qualifier);
      t = tokens.next();
      if (!")".equals(t))
	throw new XSLException("expected )", tokens.getNode());
      t = tokens.next();
    }
    else {
      if ("[".equals(t)) {
	qualifier = parseQualifiers(tokens, qualifier);
	t = tokens.next();
      }
      if (head.equals("*"))
	pattern = new WildcardPattern(qualifier);
      else
	pattern = new ElementPattern(tokens.expandElementTypeName(head), qualifier);
    }
    if (t == null)
      return pattern;
    if (t.equals("/"))
      return parseMatcher(tokens,
			  new ParentQualifier(pattern, null));
    if (t.equals("//"))
      return parseMatcher(tokens,
			  new AncestorQualifier(pattern, null));
    if (t.equals("|") || t.equals(")"))
      return pattern;
    throw new XSLException("expected / or // or |", tokens.getNode());
  }

  private static
  Qualifier parseQualifiers(PatternTokenizer tokens, Qualifier qualifier) throws XSLException {
    String qualifierToken = tokens.next();
    if (PatternTokenizer.isName(qualifierToken))
      return new ChildQualifier(tokens.expandElementTypeName(qualifierToken),
				parseMoreQualifiers(tokens, qualifier));
    if (!PatternTokenizer.isFunction(qualifierToken))
      throw new XSLException("invalid qualifier syntax", tokens.getNode());
    if ("attribute(".equals(qualifierToken))
      return parseAttributeQualifier(tokens, qualifier);
    boolean negated = false;
    if (qualifierToken != null && qualifierToken.startsWith("not-")) {
      negated = true;
      qualifierToken = qualifierToken.substring(4);
    }
    if (!")".equals(tokens.next()))
      throw new XSLException("expected \")\"", tokens.getNode());
    if ("first-of-type(".equals(qualifierToken))
      return new FirstOfTypeQualifier(negated,
				      parseMoreQualifiers(tokens, qualifier));
    if ("last-of-type(".equals(qualifierToken))
      return new LastOfTypeQualifier(negated,
				     parseMoreQualifiers(tokens, qualifier));
    if ("only-of-type(".equals(qualifierToken))
      return new OnlyOfTypeQualifier(negated,
				     parseMoreQualifiers(tokens, qualifier));
    if ("first-of-any(".equals(qualifierToken))
      return new FirstOfAnyQualifier(negated,
				     parseMoreQualifiers(tokens, qualifier));
    if ("last-of-any(".equals(qualifierToken))
      return new LastOfAnyQualifier(negated,
				    parseMoreQualifiers(tokens, qualifier));
    if ("only-of-any(".equals(qualifierToken))
      return new OnlyOfAnyQualifier(negated,
				    parseMoreQualifiers(tokens, qualifier));
    throw new XSLException("unknown qualifier", tokens.getNode());
  }

  private static
  Qualifier parseMoreQualifiers(PatternTokenizer tokens, Qualifier qualifier) throws XSLException {
    String t = tokens.next();
    if ("]".equals(t))
      return qualifier;
    if (!",".equals(t))
      throw new XSLException("expected , or ]", tokens.getNode());
    return parseQualifiers(tokens, qualifier);
  }

  private static
  Qualifier parseAttributeQualifier(PatternTokenizer tokens, Qualifier qualifier) throws XSLException {
    String nameString = tokens.next();
    if (!PatternTokenizer.isName(nameString))
      throw new XSLException("expected attribute name", tokens.getNode());
    Name attributeName = tokens.expandAttributeName(nameString);
    String t = tokens.next();
    if (!")".equals(t))
      throw new XSLException("expected \")\"", tokens.getNode());
    t = tokens.next();
    if ("=".equals(t)) {
      String attributeValue = tokens.next();
      if (attributeValue.charAt(0) != '"'
	  && attributeValue.charAt(0) != '\'')
	throw new XSLException("expected attribute value", tokens.getNode());
      attributeValue
	= attributeValue.substring(1, attributeValue.length() - 1);
      return new AttributeQualifier(attributeName,
				    attributeValue,
				    parseMoreQualifiers(tokens, qualifier));
    }
    if ("]".equals(t))
      return new AttributeQualifier(attributeName, null, qualifier);
    if (",".equals(t))
      return new AttributeQualifier(attributeName,
				    null,
				    parseQualifiers(tokens, qualifier));
    throw new XSLException("expected , or ]", tokens.getNode());
  }
}
