package com.jclark.xsl;

import java.net.URL;
import java.io.IOException;
import com.jclark.xml.parse.*;
import com.jclark.xml.parse.base.*;

public class XMLProcessorImpl extends NameTableImpl implements XMLProcessor {

  static private abstract
  class NodeImpl implements Node {
    ContainerNodeImpl parent;
    private int index;
    private NodeImpl nextSibling;
    NodeImpl() {
      this.index = 0;
      this.parent = null;
      this.nextSibling = null;
    }
    NodeImpl(int index, ContainerNodeImpl parent) {
      this.index = index;
      this.parent = parent;
      this.nextSibling = null;
      if (parent.lastChild == null)
	parent.firstChild = parent.lastChild = this;
      else {
	parent.lastChild.nextSibling = this;
	parent.lastChild = this;
      }
    }
    public Node getParent() {
      return parent;
    }
    public NodeIterator getFollowingSiblings(boolean strip) {
      if (strip)
	return new StripNodeIteratorImpl(nextSibling);
      else
	return new NodeIteratorImpl(nextSibling);
    }
    public URL getURL() {
      return parent.getURL();
    }
    public int getLineNumber() {
      return parent.getLineNumber();
    }
    boolean canStrip() {
      return false;
    }
    public Node getAttribute(Name name) {
      return null;
    }
    public int getAttributeCount() {
      return 0;
    }
    public Name getAttributeName(int i) {
      throw new IndexOutOfBoundsException();
    }
    public String getAttributeValue(Name name) {
      return null;
    }
    public int getAttributeIndex(Name name) {
      return -1;
    }
    public String getAttributeValue(int i) {
      throw new IndexOutOfBoundsException();
    }
    public Name getName() {
      return null;
    }
    public NamespacePrefixMap getNamespacePrefixMap() {
      return parent.nsMap;
    }
    public int compareTo(Node node) {
      return index - ((NodeImpl)node).index;
    }
  }

  static private
  class NodeIteratorImpl implements NodeIterator {
    private NodeImpl nextNode;

    NodeIteratorImpl(NodeImpl nextNode) {
      this.nextNode = nextNode;
    }

    public Node next() {
      NodeImpl tem = nextNode;
      if (tem != null)
	nextNode = tem.nextSibling;
      return tem;
    }
  }

  static private
  class StripNodeIteratorImpl implements NodeIterator {
    private NodeImpl nextNode;

    StripNodeIteratorImpl(NodeImpl nextNode) {
      this.nextNode = nextNode;
    }

    public Node next() {
      NodeImpl tem = nextNode;
      while (tem != null) {
	nextNode = tem.nextSibling;
	if (!tem.canStrip())
	  break;
	tem = nextNode;
      }
      return tem;
    }
  }


  static private abstract
  class ContainerNodeImpl extends NodeImpl {
    NodeImpl firstChild;
    NodeImpl lastChild;
    NamespacePrefixMap nsMap;

    ContainerNodeImpl(NamespacePrefixMap nsMap) {
      this.nsMap = nsMap;
    }

    ContainerNodeImpl(int index, ContainerNodeImpl parent) {
      super(index, parent);
      nsMap = parent.nsMap;
    }

    public NodeIterator getChildren(boolean strip) {
      if (strip)
	return new StripNodeIteratorImpl(firstChild);
      else
	return new NodeIteratorImpl(firstChild);
    }

    public String getData() {
      return null;
    }

    boolean preserveSpace() {
      return false;
    }

    public NamespacePrefixMap getNamespacePrefixMap() {
      return nsMap;
    }
  }

  static private
  class RootNodeImpl extends ContainerNodeImpl {
    private URL url;
    private DTD dtd = null;
    RootNodeImpl(URL url, NamespacePrefixMap nsMap) {
      super(nsMap);
      this.url = url;
    }
    public byte getType() {
      return Node.ROOT;
    }
    public URL getURL() {
      return url;
    }
    public int getLineNumber() {
      return 1;
    }
    void setDTD(DTD dtd) {
      this.dtd = dtd;
    }
  }

  static private
  class ElementNodeImpl extends ContainerNodeImpl {
    private Name name;
    private Object[] atts;
    private int lineNumber;
    private URL url;

    ElementNodeImpl(StartElementEvent event, int index, ContainerNodeImpl parent) throws XSLException {
      super(index, parent);
      lineNumber = event.getLocation().getLineNumber();
      url = event.getLocation().getEntityBase();
      int nAtts = event.getAttributeCount();
      if (nAtts > 0) {
	int nNsAtts = 0;
	for (int i = 0; i < nAtts; i++) {
	  String tem = event.getAttributeName(i);
	  if (tem.startsWith("xmlns")) {
	    nNsAtts++;
	    if (tem.length() == 5) {
	      String ns = event.getAttributeValue(i);
	      if (ns.length() == 0)
		nsMap = nsMap.unbindDefault();
	      else
		nsMap = nsMap.bindDefault(ns);
	    }
	    else if (tem.charAt(5) == ':')
	      nsMap = nsMap.bind(tem.substring(6),
				 event.getAttributeValue(i));
	  }
	}
	int n = nAtts - nNsAtts;
	if (n > 0) {
	  Object[] vec = new Object[n*2];
	  int j = 0;
	  for (int i = 0; i < nAtts; i++) {
	    String tem = event.getAttributeName(i);
	    if (!tem.startsWith("xmlns")) {
	      vec[j++] = nsMap.expandAttributeName(tem, this);
	      // FIXME resolve relative URL
	      vec[j++] = event.getAttributeValue(i);
	    }
	  }
	  // Assign here to avoid inconsistent state if exception
	  // is thrown.
	  atts = vec;
	}
      }
      name = nsMap.expandElementTypeName(event.getName(), this);
    }

    public Name getName() {
      return name;
    }

    public byte getType() {
      return Node.ELEMENT;
    }

    public int getAttributeCount() {
      return atts == null ? 0 : atts.length/2;
    }

    public int getAttributeIndex(Name name) {
      if (atts != null) {
	for (int i = 0; i < atts.length; i += 2)
	  if (atts[i].equals(name))
	    return i >> 1;
      }
      return -1;
    }

    public Node getAttribute(Name name) {
      int i = getAttributeIndex(name);
      if (i < 0)
	return null;
      return new AttributeNodeImpl(this.index + i + 1, this, name, getAttributeValue(i));
    }

    public String getAttributeValue(Name name) {
      if (atts != null) {
	for (int i = 0; i < atts.length; i += 2)
	  if (atts[i].equals(name))
	    return (String)atts[i + 1];
      }
      return null;
    }
    public Name getAttributeName(int i) {
      return (Name)atts[i * 2];
    }
    public String getAttributeValue(int i) {
      return (String)atts[i * 2 + 1];
    }
    public int getLineNumber() {
      return lineNumber;
    }
    public URL getURL() {
      return url;
    }
  }

  static private
  class PreserveElementNodeImpl extends ElementNodeImpl {
    PreserveElementNodeImpl(StartElementEvent event, int index, ContainerNodeImpl parent)  throws XSLException {
      super(event, index, parent);
    }
    boolean preserveSpace() {
      return true;
    }
  }

  static private
  class AttributeNodeImpl extends NodeImpl {
    private Name name;
    private String value;
    AttributeNodeImpl(int index, ContainerNodeImpl parent, Name name, String value) {
      // Don't use super(parent) because it add's this node to the children.
      this.index = index;
      this.parent = parent;
      this.name = name;
      this.value = value;
    }
    public byte getType() {
      return Node.ATTRIBUTE;
    }
    public String getData() {
      return value;
    }
    public Name getName() {
      return this.name;
    }
    public NodeIterator getChildren(boolean strip) {
      return new NodeIteratorImpl(null);
    }
  }

  static private
  class TextNodeImpl extends NodeImpl {
    private String data;
     
    public TextNodeImpl(char[] buf, int off, int len, int index, ContainerNodeImpl parent) {
      super(index, parent);
      data = new String(buf, off, len);
    }
    public byte getType() {
      return Node.TEXT;
    }
    public String getData() {
      return data;
    }
    public NodeIterator getChildren(boolean strip) {
      return new NodeIteratorImpl(null);
    }

  }

  static private
  class StripTextNodeImpl extends TextNodeImpl {
    public StripTextNodeImpl(char[] buf, int off, int len, int index, ContainerNodeImpl parent) {
      super(buf, off, len, index, parent);
    }
    boolean canStrip() {
      return true;
    }
  }


  static private
  class Builder extends ApplicationImpl {
    char[] dataBuf = new char[1024];
    int dataBufUsed = 0;
    RootNodeImpl rootNode;
    ContainerNodeImpl currentNode;
    int currentIndex = 1;

    Builder(URL url, NamespacePrefixMap nsMap) {
      currentNode = rootNode = new RootNodeImpl(url, nsMap);
    }

    public void endProlog(EndPrologEvent event) {
      rootNode.setDTD(event.getDTD());
    }

    public void startElement(StartElementEvent event) throws XSLException {
      flushData();
      ElementNodeImpl element;
      boolean preserve;
      String space = event.getAttributeValue("xml:space");
      if (space == null)
	preserve = currentNode.preserveSpace();
      else if (space.equals("default"))
	preserve = false;
      else if (space.equals("preserve"))
	preserve = true;
      else
	preserve = currentNode.preserveSpace();
      if (preserve)
	element = new PreserveElementNodeImpl(event, currentIndex++, currentNode);
      else
	element = new ElementNodeImpl(event, currentIndex++, currentNode);
      currentIndex += event.getAttributeCount();
      currentNode = element;
    }

    public void characterData(CharacterDataEvent event) {
      int need = event.getLengthMax() + dataBufUsed;
      if (need > dataBuf.length) {
	int newLength = dataBuf.length << 1;
	while (need > newLength)
	  newLength <<= 1;
	char[] tem = dataBuf;
	dataBuf = new char[newLength];
	if (dataBufUsed > 0)
	  System.arraycopy(tem, 0, dataBuf, 0, dataBufUsed);
      }
      dataBufUsed += event.copyChars(dataBuf, dataBufUsed);
    }

    public void endElement(EndElementEvent event) {
      flushData();
      currentNode = currentNode.parent;
    }

    Node getRootNode() {
      return rootNode;
    }

    private static boolean isWhitespace(char[] buf, int len) {
      for (int i = 0; i < len; i++) {
	switch (buf[i]) {
	case ' ':
	case '\n':
	case '\r':
	case '\t':
	  break;
	default:
	  return false;
	}
      }
      return true;
    }

    private final void flushData() {
      if (dataBufUsed > 0) {
	if (!currentNode.preserveSpace()
	    && isWhitespace(dataBuf, dataBufUsed))
	  new StripTextNodeImpl(dataBuf, 0, dataBufUsed, currentIndex++, currentNode);
	else
	  new TextNodeImpl(dataBuf, 0, dataBufUsed, currentIndex++, currentNode);
	dataBufUsed = 0;
      }
    }
  }

  public Node load(OpenEntity entity) throws IOException, XSLException {
    Parser parser = new ParserImpl();
    Builder builder = new Builder(entity.getBase(), getEmptyNamespacePrefixMap());
    parser.setApplication(builder);
    try {
      parser.parseDocument(entity);
    }
    catch (ApplicationException e) {
      throw (XSLException)e.getException();
    }
    return builder.getRootNode();
  }

  public Node load(URL url) throws IOException, XSLException {
    return load(new OpenEntity(url.openStream(),
			       url.toString(),
			       url,
			       null));
  }

}
